diff -uNr trunk/common/common.c trunk-new/common/common.c
--- trunk/common/common.c	Tue Jul 17 08:11:18 2007
+++ trunk-new/common/common.c	Sat Dec 22 20:53:08 2007
@@ -245,6 +245,13 @@
         else
             p->i_threads = atoi(value);
     }
+    OPT("thread-queue")
+    {
+        if( !strcmp(value, "auto") )
+            p->i_thread_queue = 0;
+        else
+            p->i_thread_queue = atoi(value);
+    }
     OPT2("deterministic", "n-deterministic")
         p->b_deterministic = atobool(value);
     OPT2("level", "level-idc")
diff -uNr trunk/common/common.h trunk-new/common/common.h
--- trunk/common/common.h	Thu Dec 20 16:24:16 2007
+++ trunk-new/common/common.h	Sat Dec 22 20:53:08 2007
@@ -233,12 +233,19 @@
 struct x264_t
 {
     /* encoder parameters */
-    x264_param_t    param;
+    x264_param_t         param;
 
-    x264_t          *thread[X264_THREAD_MAX];
-    x264_pthread_t  thread_handle;
-    int             b_thread_active;
-    int             i_thread_phase; /* which thread to use for the next frame */
+    x264_t               *thread[X264_THREAD_MAX]; /* contexts for each frame in progress */
+    x264_t               **thread_queue; /* frames that have been prepared but not yet claimed by a worker thread */
+    x264_pthread_cond_t  thread_queue_cv;
+    x264_pthread_mutex_t thread_queue_mutex;
+    x264_pthread_t       *thread_handle;
+    int                  thread_active;
+    x264_pthread_cond_t  thread_active_cv;
+    x264_pthread_mutex_t thread_active_mutex;
+    int                  thread_exit;
+    int                  b_thread_active;
+    int                  i_thread_phase; /* which thread to use for the next frame */
 
     /* bitstream output */
     struct
diff -uNr trunk/encoder/encoder.c trunk-new/encoder/encoder.c
--- trunk/encoder/encoder.c	Sat Dec 22 20:49:27 2007
+++ trunk-new/encoder/encoder.c	Sat Dec 22 20:53:08 2007
@@ -63,6 +63,34 @@
                                     x264_nal_t **pp_nal, int *pi_nal,
                                     x264_picture_t *pic_out );
 
+/* threading */
+
+static int x264_slices_write_thread( x264_t *h );
+
+#ifdef HAVE_PTHREAD
+void x264_cond_broadcast( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{
+    x264_pthread_mutex_lock( mutex );
+    *var = val;
+    x264_pthread_cond_broadcast( cv );
+    x264_pthread_mutex_unlock( mutex );
+}
+
+void x264_cond_wait( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{
+    x264_pthread_mutex_lock( mutex );
+    while( *var != val )
+        x264_pthread_cond_wait( cv, mutex );
+    x264_pthread_mutex_unlock( mutex );
+}
+
+#else
+void x264_cond_broadcast( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{}
+void x264_cond_wait( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{}
+#endif
+
 /****************************************************************************
  *
  ******************************* x264 libs **********************************
@@ -359,7 +387,12 @@
         if( h->param.i_scenecut_threshold >= 0 )
             h->param.b_pre_scenecut = 1;
 #endif
+        if( h->param.i_thread_queue == 0 )
+            h->param.i_thread_queue = h->param.i_threads;
+        h->param.i_thread_queue = x264_clip3( h->param.i_thread_queue, h->param.i_threads, X264_THREAD_MAX );
     }
+    else
+        h->param.i_thread_queue = 1;
 
     if( h->param.b_interlaced )
     {
@@ -636,7 +669,7 @@
     h->mb.i_mb_count = h->sps->i_mb_width * h->sps->i_mb_height;
 
     /* Init frames. */
-    h->frames.i_delay = h->param.i_bframe + h->param.i_threads - 1;
+    h->frames.i_delay = h->param.i_bframe + h->param.i_thread_queue - 1;
     h->frames.i_max_ref0 = h->param.i_frame_reference;
     h->frames.i_max_ref1 = h->sps->vui.i_num_reorder_frames;
     h->frames.i_max_dpb  = h->sps->vui.i_max_dec_frame_buffering;
@@ -693,17 +726,39 @@
 
     h->thread[0] = h;
     h->i_thread_num = 0;
-    for( i = 1; i < h->param.i_threads; i++ )
+    for( i = 1; i < h->param.i_thread_queue; i++ )
         h->thread[i] = x264_malloc( sizeof(x264_t) );
 
-    for( i = 0; i < h->param.i_threads; i++ )
+    if( h->param.i_threads > 1 )
+    {
+        h->thread_handle = x264_malloc( h->param.i_threads * sizeof(x264_pthread_t) );
+        h->thread_queue = x264_malloc( (h->param.i_thread_queue + 1) * sizeof(x264_t*) );
+        memset( h->thread_queue, 0, (h->param.i_thread_queue + 1) * sizeof(x264_t*) );
+        x264_pthread_cond_init( &h->thread_queue_cv, NULL );
+        x264_pthread_mutex_init( &h->thread_queue_mutex, NULL );
+    }
+
+    for( i = 0; i < h->param.i_thread_queue; i++ )
     {
+        x264_t *t = h->thread[i];
         if( i > 0 )
-            *h->thread[i] = *h;
-        h->thread[i]->fdec = x264_frame_pop_unused( h );
-        h->thread[i]->out.p_bitstream = x264_malloc( h->out.i_bitstream );
-        if( x264_macroblock_cache_init( h->thread[i] ) < 0 )
+            *t = *h;
+        t->fdec = x264_frame_pop_unused( h );
+        t->out.p_bitstream = x264_malloc( h->out.i_bitstream );
+        if( x264_macroblock_cache_init( t ) < 0 )
             return NULL;
+
+        if( h->param.i_threads > 1 )
+        {
+            x264_pthread_cond_init( &t->thread_active_cv, NULL );
+            x264_pthread_mutex_init( &t->thread_active_mutex, NULL );
+        }
+    }
+
+    if( h->param.i_threads > 1 )
+    {
+        for( i = 0; i < h->param.i_threads; i++ )
+            x264_pthread_create( &h->thread_handle[i], NULL, (void*)x264_slices_write_thread, h );
     }
 
     if( x264_ratecontrol_new( h ) < 0 )
@@ -1231,6 +1286,36 @@
     return 0;
 }
 
+static int x264_slices_write_thread( x264_t *h )
+{
+    for(;;)
+    {
+        int i_frame_size;
+        x264_t *t = NULL;
+
+        // get one frame from the queue
+        x264_pthread_mutex_lock( &h->thread_queue_mutex );
+        while( !h->thread_queue[0] && !h->thread_exit )
+            x264_pthread_cond_wait( &h->thread_queue_cv, &h->thread_queue_mutex );
+        if( h->thread_queue[0] )
+            t = (void*)x264_frame_shift( (void*)h->thread_queue );
+        x264_pthread_mutex_unlock( &h->thread_queue_mutex );
+        if( h->thread_exit )
+            return 0;
+        if( !t )
+            continue;
+
+        x264_stack_align( x264_slice_write, t );
+        i_frame_size = t->out.nal[t->out.i_nal-1].i_payload;
+        x264_fdec_filter_row( t, t->sps->i_mb_height );
+        t->out.i_frame_size = i_frame_size;
+
+        x264_cond_broadcast( &t->thread_active_cv, &t->thread_active_mutex, &t->thread_active, 0 );
+    }
+
+    return 0;
+}
+
 /****************************************************************************
  * x264_encoder_encode:
  *  XXX: i_poc   : is the poc of the current given picture
@@ -1258,7 +1343,7 @@
     if( h->param.i_threads > 1)
     {
         int i = ++h->i_thread_phase;
-        int t = h->param.i_threads;
+        int t = h->param.i_thread_queue;
         thread_current = h->thread[ i%t ];
         thread_prev    = h->thread[ (i-1)%t ];
         thread_oldest  = h->thread[ (i+1)%t ];
@@ -1302,7 +1387,7 @@
         if( h->frames.b_have_lowres )
             x264_frame_init_lowres( h, fenc );
 
-        if( h->frames.i_input <= h->frames.i_delay + 1 - h->param.i_threads )
+        if( h->frames.i_input <= h->frames.i_delay + 1 - h->param.i_thread_queue )
         {
             /* Nothing yet to encode */
             /* waiting for filling bframe buffer */
@@ -1476,8 +1561,14 @@
     /* Write frame */
     if( h->param.i_threads > 1 )
     {
-        x264_pthread_create( &h->thread_handle, NULL, (void*)x264_slices_write, h );
+        assert( h->thread_active == 0 );
+        h->thread_active = 1;
+        assert( h->b_thread_active == 0 );
         h->b_thread_active = 1;
+        x264_pthread_mutex_lock( &h->thread[0]->thread_queue_mutex );
+        x264_frame_push( (void*)h->thread_queue, (void*)h );
+        x264_pthread_cond_broadcast( &h->thread[0]->thread_queue_cv );
+        x264_pthread_mutex_unlock( &h->thread[0]->thread_queue_mutex );
     }
     else
         x264_slices_write( h );
@@ -1596,7 +1687,7 @@
 
     if( h->b_thread_active )
     {
-        x264_pthread_join( h->thread_handle, NULL );
+        x264_cond_wait( &h->thread_active_cv, &h->thread_active_mutex, &h->thread_active, 0 );
         h->b_thread_active = 0;
     }
     if( !h->out.i_nal )
@@ -1767,25 +1858,39 @@
     int64_t i_yuv_size = 3 * h->param.i_width * h->param.i_height / 2;
     int i;
 
-    for( i=0; i<h->param.i_threads; i++ )
+    if( h->param.i_threads > 1 )
     {
         // don't strictly have to wait for the other threads, but it's simpler than cancelling them
-        if( h->thread[i]->b_thread_active )
-        {
-            x264_pthread_join( h->thread[i]->thread_handle, NULL );
-            assert( h->thread[i]->fenc->i_reference_count == 1 );
-            x264_frame_delete( h->thread[i]->fenc );
+        h->thread_exit = 1;
+        x264_pthread_mutex_lock( &h->thread_queue_mutex );
+        x264_pthread_cond_broadcast( &h->thread_queue_cv );
+        x264_pthread_mutex_unlock( &h->thread_queue_mutex );
+        for( i=0; i < h->param.i_threads; i++ )
+            x264_pthread_join( h->thread_handle[i], NULL );
+        for( i=0; i < h->param.i_thread_queue; i++ )
+        {
+            x264_pthread_cond_destroy( &h->thread[i]->thread_active_cv );
+            x264_pthread_mutex_destroy( &h->thread[i]->thread_active_mutex );
+            if( h->thread[i]->b_thread_active )
+            {
+                assert( h->thread[i]->fenc->i_reference_count == 1 );
+                x264_frame_delete( h->thread[i]->fenc );
+            }
         }
+        x264_pthread_cond_destroy( &h->thread_queue_cv );
+        x264_pthread_mutex_destroy( &h->thread_queue_mutex );
+        x264_free( h->thread_handle );
+        x264_free( h->thread_queue );
     }
 
     if( h->param.i_threads > 1)
     {
         x264_t *thread_prev;
 
-        thread_prev = h->thread[ h->i_thread_phase % h->param.i_threads ];
+        thread_prev = h->thread[ h->i_thread_phase % h->param.i_thread_queue ];
         x264_thread_sync_ratecontrol( h, thread_prev, h );
         x264_thread_sync_ratecontrol( thread_prev, thread_prev, h );
-        h->i_frame = thread_prev->i_frame + 1 - h->param.i_threads;
+        h->i_frame = thread_prev->i_frame + 1 - h->param.i_thread_queue;
     }
     h->i_frame++;
 
@@ -1965,7 +2070,7 @@
     x264_cqm_delete( h );
 
     if( h->param.i_threads > 1)
-        h = h->thread[ h->i_thread_phase % h->param.i_threads ];
+        h = h->thread[ h->i_thread_phase % h->param.i_thread_queue ];
 
     /* frames */
     for( i = 0; h->frames.current[i]; i++ )
@@ -1986,7 +2091,7 @@
 
     h = h->thread[0];
 
-    for( i = h->param.i_threads - 1; i >= 0; i-- )
+    for( i = h->param.i_thread_queue - 1; i >= 0; i-- )
     {
         x264_frame_t **frame;
 
diff -uNr trunk/encoder/ratecontrol.c trunk-new/encoder/ratecontrol.c
--- trunk/encoder/ratecontrol.c	Sat Dec 22 20:50:11 2007
+++ trunk-new/encoder/ratecontrol.c	Sat Dec 22 20:53:08 2007
@@ -176,8 +176,8 @@
 
     x264_cpu_restore( h->param.cpu );
 
-    rc = h->rc = x264_malloc( h->param.i_threads * sizeof(x264_ratecontrol_t) );
-    memset( rc, 0, h->param.i_threads * sizeof(x264_ratecontrol_t) );
+    rc = h->rc = x264_malloc( h->param.i_thread_queue * sizeof(x264_ratecontrol_t) );
+    memset( rc, 0, h->param.i_thread_queue * sizeof(x264_ratecontrol_t) );
 
     rc->b_abr = h->param.rc.i_rc_method != X264_RC_CQP && !h->param.rc.b_stat_read;
     rc->b_2pass = h->param.rc.i_rc_method == X264_RC_ABR && h->param.rc.b_stat_read;
@@ -457,7 +457,7 @@
         x264_free( p );
     }
 
-    for( i=1; i<h->param.i_threads; i++ )
+    for( i=1; i<h->param.i_thread_queue; i++ )
     {
         h->thread[i]->rc = rc+i;
         rc[i] = rc[0];
@@ -1163,9 +1163,9 @@
     {
         int j = h->rc - h->thread[0]->rc;
         int i;
-        for( i=1; i<h->param.i_threads; i++ )
+        for( i=1; i<h->param.i_thread_queue; i++ )
         {
-            x264_t *t = h->thread[ (j+i)%h->param.i_threads ];
+            x264_t *t = h->thread[ (j+i)%h->param.i_thread_queue ];
             double bits = t->rc->frame_size_planned;
             if( !t->b_thread_active )
                 continue;
@@ -1316,13 +1316,13 @@
             //FIXME adjust abr_buffer based on distance to the end of the video
             int64_t diff;
 
-            if( h->fenc->i_frame < h->param.i_threads )
+            if( h->fenc->i_frame < h->param.i_thread_queue )
                 diff = total_bits + (int64_t)h->fenc->i_frame * rcc->bitrate / rcc->fps - (int64_t)rce.expected_bits;
             else
-                diff = total_bits + (int64_t)(h->param.i_threads - 1) * rcc->bitrate / rcc->fps - (int64_t)rce.expected_bits;
+                diff = total_bits + (int64_t)(h->param.i_thread_queue - 1) * rcc->bitrate / rcc->fps - (int64_t)rce.expected_bits;
             q = rce.new_qscale;
             q /= x264_clip3f((double)(abr_buffer - diff) / abr_buffer, .5, 2);
-            if( h->fenc->i_frame + 1 - h->param.i_threads >= rcc->fps && rcc->expected_bits_sum > 0)
+            if( h->fenc->i_frame + 1 - h->param.i_thread_queue >= rcc->fps && rcc->expected_bits_sum > 0)
             {
                 /* Adjust quant based on the difference between
                  * achieved and expected bitrate so far */
@@ -1368,7 +1368,7 @@
             }
             else
             {
-                int i_frame_done = h->fenc->i_frame + 1 - h->param.i_threads;
+                int i_frame_done = h->fenc->i_frame + 1 - h->param.i_thread_queue;
 
                 q = get_qscale( h, &rce, rcc->wanted_bits_window / rcc->cplxr_sum, h->fenc->i_frame );
 
diff -uNr trunk/x264.c trunk-new/x264.c
--- trunk/x264.c	Tue Jul 17 08:11:18 2007
+++ trunk-new/x264.c	Sat Dec 22 20:53:08 2007
@@ -304,6 +304,7 @@
     H0( "      --no-psnr               Disable PSNR computation\n" );
     H0( "      --no-ssim               Disable SSIM computation\n" );
     H0( "      --threads <integer>     Parallel encoding\n" );
+    H1( "      --thread-queue <integer> Number of delay frames for thread sync\n" );
     H0( "      --thread-input          Run Avisynth in its own thread\n" );
     H1( "      --non-deterministic     Slightly improve quality of SMP, at the cost of repeatability\n" );
     H1( "      --no-asm                Disable all CPU optimizations\n" );
@@ -425,6 +426,7 @@
             { "zones",   required_argument, NULL, 0 },
             { "qpfile",  required_argument, NULL, OPT_QPFILE },
             { "threads", required_argument, NULL, 0 },
+            { "thread-queue", required_argument, NULL, 0 },
             { "thread-input", no_argument,  NULL, OPT_THREAD_INPUT },
             { "non-deterministic", no_argument, NULL, 0 },
             { "no-psnr", no_argument,       NULL, 0 },
diff -uNr trunk/x264.h trunk-new/x264.h
--- trunk/x264.h	Tue Nov 20 02:57:28 2007
+++ trunk-new/x264.h	Sat Dec 22 20:53:08 2007
@@ -142,6 +142,7 @@
     /* CPU flags */
     unsigned int cpu;
     int         i_threads;       /* encode multiple frames in parallel */
+    int         i_thread_queue;  /* number of frames to prepare in advance (>= i_threads) */
     int         b_deterministic; /* whether to allow non-deterministic optimizations when threaded */
 
     /* Video Properties */